#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdint.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "io.h"
#include "buffer.h"
#include "hal-proto.h"
#include "hal-cmd-parser.h"
#include "hal-peer.h"

#define warn(fmt, arg...) fprintf(stderr, "warn: %s: "fmt CRLF, \
					__FUNCTION__,  ##arg)

#define	HAL_PEER_EVENT_HANDLER(peer, event)		\
do {							\
	if (peer->event_handler != NULL)		\
		peer->event_handler(peer, event); 	\
} while (0) 
		
struct hal_peer_base *hal_peer_base_new()
{
	struct hal_peer_base *peer;

	peer = xmalloc0(sizeof(*peer));

	peer->wfd = -1;
	peer->rfd = -1;

	return peer;
}

void hal_peer_base_free(struct hal_peer_base **peer)
{
	return_if_fail(peer != NULL);
	return_if_fail((*peer) != NULL);

	xfree(*peer);
	*peer = NULL;
}

void hal_peer_base_init(struct hal_peer_base *peer, struct io_context *io,
			int wfd, int rfd, hal_peer_event_callback_t event_handler)
{
	return_if_fail(peer != NULL);
	return_if_fail(wfd >= 0);
	return_if_fail(rfd >= 0);

	peer->wfd = wfd;
	peer->rfd = rfd;
	peer->wb = buffer_new();
	peer->rb = buffer_new();
	peer->io = io;
	peer->event_handler = event_handler;

	if (rfd != wfd) {
		io_add(io, wfd, EVENT_WRITE, hal_peer_event_default, peer);
		io_add(io, rfd, EVENT_READ, hal_peer_event_default, peer);
	} else {
		io_add(io, rfd, EVENT_READ, hal_peer_event_default, peer);
	}
}

void hal_peer_base_deinit(struct hal_peer_base *peer)
{
	return_if_fail(peer != NULL);

	if (peer->rfd != peer->wfd) {
		io_del(peer->io, peer->wfd);
		io_del(peer->io, peer->rfd);
		close(peer->wfd);
		close(peer->rfd);
	} else {
		io_del(peer->io, peer->rfd);
		close(peer->rfd);
	}

	peer->wfd = -1;
	peer->rfd = -1;
	peer->io = NULL;
	peer->event_handler = NULL;

	buffer_free(peer->wb);
	buffer_free(peer->rb);
	peer->wb = NULL;
	peer->rb = NULL;
}

void hal_peer_event_default(int fd, short events, void *opaque)
{
	struct hal_peer_base *peer;
	int res;

	peer = (struct hal_peer_base *)opaque;

	if (events & EVENT_ERROR) {
		warn("fd=%d error", fd);
		HAL_PEER_EVENT_HANDLER(peer, HAL_PEER_EVENT_ERROR);
		return;
	}

	if (events & EVENT_TIMEOUT) {
		warn("timeout");
		HAL_PEER_EVENT_HANDLER(peer, HAL_PEER_EVENT_TIMEOUT);
		return;
	}

	if (events & EVENT_READ) {
		res = buffer_read_fd(peer->rfd, peer->rb, 4096);
		if (res == 0) {
			HAL_PEER_EVENT_HANDLER(peer, HAL_PEER_EVENT_ERROR);
		} else if (res == -1 && (errno != EAGAIN && errno != EINTR &&
					 errno != EWOULDBLOCK)) {
			HAL_PEER_EVENT_HANDLER(peer, HAL_PEER_EVENT_ERROR);
		} else if (res > 0) {
			hal_cmd_parse_packet(peer->rb, peer);
			HAL_PEER_EVENT_HANDLER(peer, HAL_PEER_EVENT_READ);
		}
	}

	if (events & EVENT_WRITE) {
		if (buffer_size(peer->wb) != 0) {
			res = buffer_write_fd(peer->wfd, peer->wb);
			if (res == -1 && (errno != EAGAIN && errno != EINTR &&
					  errno != EWOULDBLOCK)) {
				HAL_PEER_EVENT_HANDLER(peer, HAL_PEER_EVENT_ERROR);
			} else if (res > 0) {
				HAL_PEER_EVENT_HANDLER(peer, HAL_PEER_EVENT_WRITE);
			}
		}
		/* This is not a pipe. */
		if (peer->wfd == peer->rfd) {
			if (buffer_size(peer->wb) == 0)
				io_change(peer->io, peer->wfd, EVENT_READ);
			else
				io_change(peer->io, peer->wfd, EVENT_READ|EVENT_WRITE);	
		} 
	}
}

void hal_peer_send_default(struct hal_peer_base *peer)
{
	if (peer->wfd == peer->rfd &&
	    buffer_size(peer->wb) != 0)
		io_change(peer->io, peer->wfd, EVENT_READ|EVENT_WRITE);
}

